/* San Diego Road Runner login program for Linux/BSDI/FreeBSD
 * 
 * Revised 31 Dec 1997 to clean up initializations of sockaddr_in
 * structures to try to fix garbage local addresses in getsockname() calls
 * for certain shared versions of the Linux C library. Not yet verified.
 *
 * Revised 22 Aug 1997 to add null statement to default label
 * Revised: 19 July 1997 to update address of 7507 router
 *
 * Revised: 15 Jun 1997 P. Karn, karn@ka9q.ampr.org,
 * to automate determination of TAS IP address, to accept a TCP RST
 * from the test host as confirmation that we're still logged in,
 * to use SDRR's Cisco router as the test host instead of one of the
 * proxy servers, and to ignore SIGHUP and SIGPIPE.
 *
 * NB! This program is specifically designed to work with the *San Diego*
 * Road Runner network. RR networks in other cities use different IP
 * addresses and may use different login protocols. This program is
 * NOT designed to work outside San Diego. It will automatically detect
 * if this is the case and exit.
 *
 * This program now automatically determines the hub you're on
 * and the correct IP address of the TAS to use. If you wish to override
 * this, define the AUTHHOST macro below with the correct TAS IP address.
 *
 * See <http://people.qualcomm.com/karn/rr/rrarch.html> for a listing
 * of hubs and TAS IP addresses.
 */

/* Modified for OS/2 by Larkin Lowrey (llowrey@NuclearWinter.com) 1-15-97
 * OS/2 Road Runner info page: http://larkin.NuclearWinter.com/rros2.html
 *
 * This is a preliminary release. It works for me but I make no guarantees.
 * This has only been tested with WARP 4 with the latest MPTS fixpack installed.
 */

/*#define AUTHHOST "204.210.7.25"*/ /* Hub 6 (University City) */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <netdb.h>
#include <time.h>
//#include <syslog.h>
#include <signal.h>
#include <errno.h>
#define __os2__
#if defined(__os2__) || defined(__bsdi__) || defined(__FreeBSD__) || defined(__NEXT__) || defined(__sun__) || defined(sun)
#include <sys/socket.h>
#include <netinet/in.h>
#elif defined(__linux__)
#include <linux/socket.h>
#include <linux/in.h>
#endif

int rdmsg(FILE *fp,char *buffer,int size);
void sndmsg(int s,char *buffer);

/* Parameters specific to the San Diego Road Runner network */
#define	SANRRIP		"204.210.0.0"	/* Base of IP address block */
#define SANRRIPMASK	"255.255.192.0"	/* Subnet mask for same */
#define	SUBNET(x)	((x & 0x00003f00) >> 8)

/* Convert subnet number (third field in IP address) to hub number */
int subnettohub[] = {
0,0,0,0,0,0,0,0,	/*   0-7: Internal use/unused */
7,			/*     8: Ware Ct Test */
2,2,2,			/*  9-11: Mira Mesa */
1,1,1,1,1,1,1,		/* 12-18: Pacific Beach */
3,3,3,			/* 19-21: Scripps Ranch */
8,8,8,8,8,		/* 22-26: La Jolla */
4,4,4,4,		/* 27-30: Rancho Penasquitos */
5,5,5,5,		/* 31-34: North Clairemont */
6,6,6,6,		/* 35-38: University City/UTC */
15,15,15,		/* 39-41: Carmel Valley */
10,10,10,10,		/* 42-45: South Clairemont */
9,9,9,9,		/* 46-49: Kearney Mesa */
11,11,			/* 50-51: Rancho Bernardo West */
12,12,12,12,		/* 52-55: Rancho Bernardo East */
13,13,			/* 56-57: Tierrasanta */
0,0,0,0,0,0,		/* 58-63: Internal use/unused */
};

/* Convert hub number to region name and TAS address */
struct {
	char *tasaddr;
	char *name;
} tastable[] = {
"",			"",
"204.210.7.5",		"Pacific Beach",
"204.210.7.9",		"Mira Mesa",
"204.210.7.13",		"Scripps Ranch",
"204.210.7.17",		"Rancho Penasquitos",
"204.210.7.21",		"North Clairemont",
"204.210.7.25",		"University City",
"204.210.7.29",		"Ware Ct. Test",
"204.210.7.33",		"La Jolla",
"204.210.7.37",		"Kearney Mesa",
"204.210.7.41",		"South Clairemont",
"204.210.7.45",		"Rancho Bernardo West",
"204.210.7.49",		"Rancho Bernardo East",
"204.210.7.53",		"Tierrasanta",
"",			"unassigned",
"204.210.7.57",		"Carmel Valley",
};

/* Port numbers for Toshiba authentication server */
#define TCPPORT 7283
#define UDPPORT 6284
#define L_UDPPORT 6285

/* Use the MCI interface on the Cisco 7507 as the login test host
 * This is the nearest interface we can reach with a TCP packet
 * to an unused port without being blocked by a RR firewall
 */
//#define TESTHOST "166.48.172.14"
#define TESTHOST "time-inc-new-media.Bloomington.mci.net"
/*#define TESTHOST "204.70.252.106" */
#define TESTPORT 8080

int Verbose = 0;

main(int argc,char *argv[])
{
	int udp_s,tcp_s,r,lsize,len,c;
	char logname[10],password[10];
	struct sockaddr_in tas,local,server;
	char wbuffer[512];
	char rbuffer[100],*cp;
	int hub;
	char *authhost = NULL;
	char *testhost = TESTHOST;
	struct hostent *hp;
	FILE *net;
	extern char *optarg;
	extern int optind;
	/* The default timeout should probably be bigger, now that the
	 * TAS timeout is 20 minutes.
	 */
	int timeout = 60;
	int logoff_first = 0;
	time_t t;
	unsigned long ipbase,netmask;
	enum { down, up} state = down;

	server.sin_addr.s_addr = 0;

	while((c = getopt(argc,argv,"lt:h:dvp:")) != EOF){
		switch(c){
		case 'l': /* Send logoff packet before logging on */
			logoff_first++;
			break;
		case 't':
			timeout = atoi(optarg);
			break;
		case 'h':
			authhost = optarg;
			break;
		case 'v':			
		case 'd':
			Verbose++;
			break;
		case 'p': /* test server, for testing login state */
			testhost = optarg;
			break;
		default: ;
		}
	}
	/* Determine local RR IP address by connecting a UDP socket to
	 * the base of the RR IP address space and looking at the local
	 * address. NB! This assumes the DHCP client has obtained a
	 * local IP address and routing is set up correctly!
	 */
	if((udp_s = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP)) == -1){
		perror("can't create udp socket");
		exit(1);
	}
	tas.sin_family = AF_INET;
	tas.sin_port = htons(UDPPORT);
    tas.sin_addr.s_addr=inet_addr(SANRRIP);

	if(connect(udp_s,(void *)&tas,sizeof(tas)) == -1){
		perror("udp connect");
		exit(1);
	}
	lsize = sizeof(local);
	getsockname(udp_s,(void *)&local,&lsize);
	close(udp_s);

	/* Verify that we're on the San Diego RR system and that
	 * DHCP and routing are properly set up
	 */
    ipbase=inet_addr(SANRRIP);
    netmask=inet_addr(SANRRIPMASK);
	if(Verbose){
		printf("Local IP address: %s\n",
         inet_ntoa(*(struct in_addr *)&local.sin_addr.s_addr));
	}
	if((local.sin_addr.s_addr & netmask) != ipbase){
		printf("Error: This address does not belong to the San Diego Road Runner system.\n");
		printf("Check that your DHCP client has obtained an IP address and has\n");
		printf("installed the correct routing entries\n");
		exit(1);
	}
	hub = subnettohub[SUBNET(ntohl(local.sin_addr.s_addr))];
	if(authhost == NULL){
#ifdef	AUTHHOST
		authhost = AUTHHOST;
#else
		authhost = tastable[hub].tasaddr;
#endif
	}
	if(Verbose){
		printf("Hub #%d (%s); TAS IP addr: %s\n",
		 hub,tastable[hub].name,authhost);
	}
	/* If login name and/or password were not given on the command
	 * line, prompt for them here
	 */
	if(argc <= optind){
		printf("Enter roadrunner login name: ");
		fgets(logname,sizeof(logname),stdin);
	} else {
		strcpy(logname,argv[optind]);
	}
	if((cp = strchr(logname,'\n')) != NULL)
		*cp = '\0';

	if(argc <= optind+1){
		printf("Enter password: ");
		fgets(password,sizeof(password),stdin);
	} else {
		strcpy(password,argv[optind+1]);
	}
	if((cp = strchr(password,'\n')) != NULL)
		*cp = '\0';

	/* If not running in verbose (debug) mode, fork off a child
	 * and have the parent immediately exit (this is standard
	 * practice for UNIX system daemons)
	 */
	if(!Verbose && fork() != 0)
		exit(0);

	signal(SIGHUP,SIG_IGN);
	signal(SIGPIPE,SIG_IGN);

	/* Create UDP socket for keepalive messages */
	if((udp_s = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP)) == -1){
		perror("UDP socket call failed");
		exit(1);
	}
	local.sin_family = AF_INET;
	local.sin_port = htons(L_UDPPORT);
	/* local.sin_addr.s_addr still set from earlier getsockname() */
	bind(udp_s,(void *)&local,sizeof(local));

	tas.sin_family = AF_INET;
	tas.sin_port = htons(UDPPORT);
    tas.sin_addr.s_addr=inet_addr(authhost);
	if(connect(udp_s,(void *)&tas,sizeof(tas)) == -1){
		perror("UDP connect to tas failed");
		close(udp_s);
		exit(1);
	}
	/* tas.sin_family, tas.sin_addr.s_addr still set from above */
	tas.sin_port = htons(TCPPORT);

retry:	/* Open TCP connection to authentication server */
	if((tcp_s = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP)) == -1){
		perror("TCP socket call failed");
		sleep(timeout);
		goto retry;
	}
	if(connect(tcp_s,(void *)&tas,sizeof(tas)) == -1){
		perror("TCP connect to tas failed");
		close(tcp_s);
		sleep(timeout);
		goto retry;
	}
	if(logoff_first){
		if(Verbose)
			printf("Sending UDP logoff msg...\n");

		/* Send UDP logout message to ensure we're not already logged in */
		sprintf(wbuffer,"00,01,0000,00000038,%-16s,",
           inet_ntoa(*(struct in_addr *)&local.sin_addr.s_addr));

		sndmsg(udp_s,wbuffer);
		sleep(3);	/* Wait for response to come back */
	}
	/* Send TCP login sequence */
	sprintf(wbuffer,"01,01,0000,00000065,%-8s,%-8s,%-16s,%-8s,",
       logname,password,inet_ntoa(*(struct in_addr *)&local.sin_addr.s_addr),"WIN-95");
	if(Verbose)
		printf("sending TCP login msg...\n");
	sndmsg(tcp_s,wbuffer);

	net = fdopen(tcp_s,"r+");	/* Use stdio for receive */

	rdmsg(net,rbuffer,sizeof(rbuffer));

#ifdef notdef
	sprintf(wbuffer,"02,03,0000,00000075,%-8s,1 ,%-32s,%-8s,",
	   os,exec,version);
	sprintf(wbuffer,"02,03,0000,00000201,WIN-95  ,4 ,rrie95.exe                      ,1.0.0.2 ,login.new                       ,1.2.0.0 ,readme.txt                      ,1.2.0.0 ,modini.exe                      ,1.0.0.2 ,");

	sprintf(wbuffer,"02,03,0000,00000033,WIN-95  ,0 ,");
#endif
	sprintf(wbuffer,"02,03,0000,00000075,Linux   ,0 ,");

	sndmsg(tcp_s,wbuffer);

	rdmsg(net,rbuffer,sizeof(rbuffer));

	sprintf(wbuffer,"03,02,0000,00000021,");
	sndmsg(tcp_s,wbuffer);

	rdmsg(net,rbuffer,sizeof(rbuffer));

	fclose(net);	/* Done with the TCP connection */

	/* Using DNS should be OK now */
	if(server.sin_addr.s_addr == 0){	 /* Do it only once */
		if((hp = gethostbyname(testhost)) == NULL)
        {
            printf("**%s\n",testhost);
            herror("Can't resolve server host");
        }
		server.sin_family = AF_INET;
		server.sin_port = htons(TESTPORT);
		memcpy(&server.sin_addr.s_addr,hp->h_addr_list[0],4);
		if(Verbose)
			printf("server host: %s (%s)\n",testhost,
              inet_ntoa(*(struct in_addr *)&server.sin_addr.s_addr));
	}
	/* Send keepalives forever */
	if(Verbose)
		printf("Sending UDP keepalives...\n");
	for(;;){
		sprintf(wbuffer,"99,03,0000,00000038,%-16s,",
           inet_ntoa(*(struct in_addr *)&local.sin_addr.s_addr));
		sndmsg(udp_s,wbuffer);

		/* Connect to head end server to verify
		 * that we're still logged in
		 */
		if((tcp_s = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP)) == -1){
			perror("TCP socket call failed");
			sleep(timeout);
			continue;
		}
		r = connect(tcp_s,(void *)&server,sizeof(server));
		close(tcp_s);
		if(r == -1 && errno != ECONNREFUSED){
			if(Verbose)
				perror("server test failed");
			if(state == up){
				state = down;
				time(&t);
                printf("RR connection down at %s",ctime(&t));
			}
			goto retry;
		}
		if(Verbose)
			printf("server check OK\n");

		/* Either the connect succeeded, or the server
		 * refused it. Either way we know we're reaching it
		 */
		if(state == down){
			state = up;
			time(&t);
            printf("RR connection up at %s",ctime(&t));
		}
		sleep(timeout);
	}
}
/* Read a message from the stream, terminating after a null has been
 * transferred or when the size of the buffer has been reached
 */
int
rdmsg(FILE *fp,char *buffer,int size)
{
	char *cp;
	int count = 0;
	int c;

	cp = buffer;
	while((c = getc(fp)) != EOF && count < size){
		*cp++ = c;
		if(c == '\0')
			break;
	}
	if(Verbose)
		printf("rx: %s\n",buffer);

	return count;
}
void
sndmsg(int s,char *buffer)
{
	if(Verbose)
		printf("tx: %s\n",buffer);
	write(s,buffer,strlen(buffer)+1);
}
